import uuid

from typing import Optional

from sqlalchemy import Boolean, Column, ForeignKey, Integer, Unicode, func, select
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import Session, relationship

from .meta import Base


class AccountBase(Base):
    __tablename__ = "accounts"

    id = Column("id", UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    code = Column("code", Integer, nullable=False)
    name = Column("name", Unicode(255), unique=True, nullable=False)
    type_id = Column("type", Integer, ForeignKey("account_types.id"), nullable=False)
    account_type = Column("account_type", Unicode(50), nullable=False)
    is_starred = Column("is_starred", Boolean, nullable=False)
    is_active = Column("is_active", Boolean, nullable=False)
    is_reconcilable = Column("is_reconcilable", Boolean, nullable=False)
    cost_centre_id = Column(
        "cost_centre_id",
        UUID(as_uuid=True),
        ForeignKey("cost_centres.id"),
        nullable=False,
    )
    is_fixture = Column("is_fixture", Boolean, nullable=False)

    __mapper_args__ = {"polymorphic_on": account_type}

    type_ = relationship("AccountType")

    journals = relationship("Journal", back_populates="account")
    cost_centre = relationship("CostCentre", back_populates="accounts")
    products = relationship("Product", back_populates="account")

    rate_contracts = relationship("RateContract", back_populates="vendor")

    @property
    def __name__(self):
        return self.name

    def __init__(
        self,
        code=None,
        name=None,
        type_id=None,
        is_starred=None,
        is_active=None,
        is_reconcilable=False,
        cost_centre_id=None,
        id_=None,
        is_fixture=False,
    ):
        self.code = code
        self.name = name
        self.type_id = type_id
        self.is_starred = is_starred
        self.is_active = is_active
        self.is_reconcilable = is_reconcilable
        self.cost_centre_id = cost_centre_id
        self.id = id_
        self.is_fixture = is_fixture

    @classmethod
    def query(cls, q: str, type_: int, reconcilable: Optional[bool], active: Optional[bool], db: Session):
        query_ = select(cls)
        if type_ is not None:
            if not isinstance(type_, int):
                type_ = int(type_)
            query_ = query_.where(cls.type_id == type_)
        if reconcilable is not None:
            query_ = query_.where(cls.is_reconcilable == reconcilable)
        if active is not None:
            query_ = query_.where(cls.is_active == active)
        if q is not None:
            for item in q.split():
                query_ = query_.where(cls.name.ilike(f"%{item}%"))
        return db.execute(query_.order_by(cls.name)).scalars().all()

    def create(self, db: Session):
        self.code = db.execute(
            select(func.coalesce(func.max(AccountBase.code), 0) + 1).where(AccountBase.type_id == self.type_id)
        ).scalar_one()
        db.add(self)
        return self

    def can_delete(self, advanced_delete: bool):
        if self.is_fixture:
            return False, f"{self.name} is a fixture and cannot be edited or deleted."
        if self.is_active:
            return False, "Account is active"
        if len(self.journals) > 0 and not advanced_delete:
            return False, "Account has journal entries"
        return True, ""

    @classmethod
    def get_code(cls, type_: int, db: Session):
        return db.execute(
            select(func.coalesce(func.max(AccountBase.code), 0) + 1).where(AccountBase.type_id == type_)
        ).scalar_one()

    @classmethod
    def all_purchases(cls):
        return uuid.UUID("240dd899-c413-854c-a7eb-67a29d154490")

    @classmethod
    def cash_in_hand(cls):
        return {"id": "ed2341bb-80b8-9649-90db-f9aaca183bb3", "name": "Cash in Hand"}

    @classmethod
    def local_purchase(cls):
        return {"id": "d2b75912-505f-2548-9093-466dfff6a0f9", "name": "Local Purchase"}

    @classmethod
    def salary(cls):
        return {"id": "5c2b54d0-c174-004d-a0d5-92cdaadcefa7", "name": "Staff Salary"}

    @classmethod
    def salary_id(cls):
        return uuid.UUID("5c2b54d0-c174-004d-a0d5-92cdaadcefa7")

    @classmethod
    def incentive(cls):
        return {"id": "b7eff754-e8ba-e047-ab06-9132c15c7640", "name": "Incentives"}

    @classmethod
    def incentive_id(cls):
        return uuid.UUID("b7eff754-e8ba-e047-ab06-9132c15c7640")

    @classmethod
    def esi_pf_expense(cls):
        return uuid.UUID("d2a1a286-e900-764b-a1a5-9f4b00dbb940")

    @classmethod
    def esi_pf_payable(cls):
        return uuid.UUID("42277912-cc18-854b-b134-9f4b00dba419")

    @classmethod
    def suspense(cls):
        return uuid.UUID("3854e317-6f3b-5142-ab26-9c44d4cddd08")
