import uuid

from datetime import date

from sqlalchemy import Column, Date, ForeignKey, Numeric, select
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import Session, contains_eager, relationship

from .meta import Base
from .product import Product
from .stock_keeping_unit import StockKeepingUnit


class Batch(Base):
    __tablename__ = "batches"

    id = Column("id", UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column("name", Date, nullable=False)
    sku_id = Column("sku_id", UUID(as_uuid=True), ForeignKey("stock_keeping_units.id"), nullable=False)
    quantity_remaining = Column("quantity_remaining", Numeric(precision=15, scale=2), nullable=False)
    rate = Column("rate", Numeric(precision=15, scale=2), nullable=False)
    tax = Column("tax", Numeric(precision=15, scale=5), nullable=False)
    discount = Column("discount", Numeric(precision=15, scale=5), nullable=False)

    inventories = relationship("Inventory", back_populates="batch")
    sku = relationship("StockKeepingUnit", back_populates="batches")

    def __init__(
        self,
        name=None,
        sku_id=None,
        quantity_remaining=None,
        rate=None,
        tax=None,
        discount=None,
        sku=None,
    ):
        self.name = name
        self.sku_id = sku_id
        self.quantity_remaining = quantity_remaining
        self.rate = rate
        self.tax = tax
        self.discount = discount
        if sku is None:
            self.sku_id = sku_id
        else:
            self.sku_id = sku.id
            self.sku = sku

    def amount(self):
        return self.quantity_remaining * self.rate * (1 + self.tax) * (1 - self.discount)

    @classmethod
    def list(cls, q: str, include_nil: bool, date_: date, db: Session):
        query = (
            select(cls)
            .join(cls.sku)
            .join(StockKeepingUnit.product)
            .options(contains_eager(cls.sku).contains_eager(StockKeepingUnit.product))
        )
        if not include_nil:
            query = query.where(cls.quantity_remaining > 0)
        if date_ is not None:
            query = query.where(cls.name <= date_)
        if q is not None:
            for item in q.split():
                query = query.where(Product.name.ilike(f"%{item}%"))
        return db.execute(query.order_by(Product.name).order_by(cls.name)).scalars().all()

    @classmethod
    def suspense(cls):
        return uuid.UUID("a955790e-93cf-493c-a816-c7d92b127383")
