import re
import uuid

from datetime import date, timedelta
from decimal import Decimal
from typing import Dict, List, Optional, Tuple

from sqlalchemy import or_, select
from sqlalchemy.orm import Session

from ..models.account_type import AccountType
from ..models.db_setting import DbSetting
from ..models.setting_type import SettingType
from ..models.voucher_type import VoucherType
from ..schemas.settings import (
    AccountTypesSelected,
    LockDate,
    LockInformation,
    VoucherTypesSelected,
)


def get_lock_info(
    voucher_dates: List[date], voucher_type: VoucherType, account_types: List[int], db: Session
) -> Tuple[bool, str]:
    date_ = date.today()
    start: Optional[date]
    finish: Optional[date]
    data: List = (
        db.execute(
            select(DbSetting.data).where(
                DbSetting.setting_type == SettingType.VOUCHER_LOCK,
                or_(
                    DbSetting.valid_from == None,  # noqa: E711
                    DbSetting.valid_from <= date_,
                ),
                or_(
                    DbSetting.valid_till == None,  # noqa: E711
                    DbSetting.valid_till >= date_,
                ),
            )
        )
        .scalars()
        .all()
    )
    data = sorted(data, key=lambda d: d["index"], reverse=True)
    for it in data:
        li: LockInformation = LockInformation(
            voucherTypes=[VoucherTypesSelected(id=vt["id_"], name=vt["name"]) for vt in it["voucher_types"]],
            accountTypes=[AccountTypesSelected(id=at["id_"], name=at["name"]) for at in it["account_types"]],
            start=LockDate(days=it["start"]["days"], date=it["start"]["date_"]),
            finish=LockDate(days=it["finish"]["days"], date=it["finish"]["date_"]),
            index=it["index"],
        )
        # Data format
        if len(li.voucher_types) != 0 and voucher_type not in [vt.id_ for vt in li.voucher_types]:
            continue
        if len(li.account_types) != 0 and len(set(account_types) & set([at.id_ for at in li.account_types])) == 0:
            continue

        if li.start.days is not None:
            start = date.today() - timedelta(days=li.start.days)
            if sum(1 for v in voucher_dates if v < start):
                return False, f"Vouchers before {start.strftime('%d-%b-%Y')} have been locked."
        if li.start.date_ is not None:
            start = li.start.date_
            if sum(1 for v in voucher_dates if v < start):
                return False, f"Vouchers before {start.strftime('%d-%b-%Y')} have been locked."

        if li.finish.days is not None:
            finish = date.today() + timedelta(days=li.finish.days)
            if sum(1 for v in voucher_dates if v > finish):
                return False, f"Vouchers after {finish.strftime('%d-%b-%Y')} have been locked."
        if li.finish.date_ is not None:
            finish = li.finish.date_
            if sum(1 for v in voucher_dates if v > finish):
                return False, f"Vouchers after {finish.strftime('%d-%b-%Y')} have been locked."

        return True, "Voucher allowed"
    return False, "Default Fallthrough"


def to_uuid(value):
    p = re.compile("^[A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}$")
    return uuid.UUID(value) if p.match(value) else None


def to_decimal(value, default=0):
    import re

    _parser = re.compile(
        r"""        # A numeric string consists of:
        (?P<sign>[-+])?              # an optional sign, followed by either...
        (
            (?=\d|\.\d)              # ...a number (with at least one digit)
            (?P<int>\d*)             # having a (possibly empty) integer part
            (\.(?P<frac>\d*))?       # followed by an optional fractional part
        |
            (?P<signal>s)?           # ...an (optionally signaling)
            NaN                      # NaN
            (?P<diag>\d*)            # with (possibly empty) diagnostic info.
        )
        \Z
    """,
        re.VERBOSE | re.IGNORECASE,
    ).match
    return Decimal(value) if _parser(value.strip()) is not None else default
