import uuid

from datetime import datetime
from decimal import Decimal
from typing import List, Optional

import brewman.schemas.product as schemas

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy import delete, desc, func, or_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session, contains_eager

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.account import Account
from ..models.batch import Batch
from ..models.product import Product
from ..models.rate_contract import RateContract
from ..models.rate_contract_item import RateContractItem
from ..models.stock_keeping_unit import StockKeepingUnit
from ..schemas.user import UserToken


router = APIRouter()


@router.post("", response_model=schemas.Product)
def save(
    data: schemas.ProductIn,
    user: UserToken = Security(get_user, scopes=["products"]),
) -> schemas.Product:
    try:
        with SessionFuture() as db:
            item = Product(
                name=data.name,
                product_group_id=data.product_group.id_,
                account_id=Account.all_purchases(),
                is_active=data.is_active,
                is_purchased=data.is_purchased,
                is_sold=data.is_sold,
            ).create(db)
            if len([s for s in data.skus if s.is_default is True]) != 1:
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Exactly one default sku is needed"
                )
            if len(set([s.fraction_units for s in data.skus])) != 1:
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail="All skus need to have the same fraction unit",
                )
            for sku in data.skus:
                db.add(
                    StockKeepingUnit(
                        is_default=sku.is_default,
                        units=sku.units,
                        fraction=round(sku.fraction, 5),
                        fraction_units=sku.fraction_units,
                        product_yield=round(sku.product_yield, 5),
                        price=round(sku.price, 2),
                        sale_price=round(sku.sale_price, 2),
                        product=item,
                    )
                )
            db.commit()
            return product_info(item)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.put("/{id_}", response_model=schemas.Product)
def update_route(
    id_: uuid.UUID,
    data: schemas.ProductIn,
    user: UserToken = Security(get_user, scopes=["products"]),
) -> schemas.Product:
    try:
        with SessionFuture() as db:
            item: Product = db.execute(select(Product).where(Product.id == id_)).scalar_one()
            if item.is_fixture:
                raise HTTPException(
                    status_code=status.HTTP_423_LOCKED,
                    detail=f"{item.name} is a fixture and cannot be edited or deleted.",
                )
            item.name = data.name
            item.product_group_id = data.product_group.id_
            item.account_id = Account.all_purchases()
            item.is_active = data.is_active
            item.is_purchased = data.is_purchased
            item.is_sold = data.is_sold
            if len([sku for sku in data.skus if sku.is_default is True]) != 1:
                raise HTTPException(
                    status_code=status.HTTP_423_LOCKED,
                    detail="There needs to be exactly 1 default SKU",
                )
            if len(set([s.fraction_units for s in data.skus])) != 1:
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail="All skus need to have the same fraction unit",
                )
            default_sku = next(s.units for s in data.skus if s.is_default is True)
            for i in range(len(item.skus), 0, -1):
                sku = item.skus[i - 1]
                index = next((idx for (idx, d) in enumerate(data.skus) if d.id_ == sku.id), None)
                if index is not None:
                    new_sku = data.skus.pop(index)
                    sku.is_default = False
                    sku.units = new_sku.units
                    sku.fraction = round(new_sku.fraction, 5)
                    sku.fraction_units = new_sku.fraction_units
                    sku.product_yield = round(new_sku.product_yield, 5)
                    sku.price = round(new_sku.price, 2)
                    sku.sale_price = round(new_sku.sale_price, 2)
                else:
                    count: Decimal = db.execute(select(func.count()).where(Batch.sku_id == sku.id)).scalar_one()
                    if count > 0:
                        raise HTTPException(
                            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            detail=f"SKU '{sku.units}' has entries and cannot be deleted.",
                        )
                    item.skus.remove(sku)
                    db.delete(sku)
            for sku in data.skus:
                new_sku = StockKeepingUnit(
                    is_default=False,
                    units=sku.units,
                    fraction=round(sku.fraction, 5),
                    fraction_units=sku.fraction_units,
                    product_yield=round(sku.product_yield, 5),
                    price=round(sku.price, 2),
                    sale_price=round(sku.sale_price, 2),
                    product=item,
                )
                db.add(new_sku)
                item.skus.append(new_sku)
            db.flush()
            default_sku = next(s for s in item.skus if s.units == default_sku)
            default_sku.is_default = True
            db.commit()
            return product_info(item)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.delete("/{id_}", response_model=schemas.ProductBlank)
def delete_route(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["products"]),
) -> schemas.ProductBlank:
    with SessionFuture() as db:
        item: Product = db.execute(select(Product).where(Product.id == id_)).scalar_one()
        if item.is_fixture:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"{item.name} is a fixture and cannot be edited or deleted.",
            )
        count: Decimal = db.execute(
            select(func.count()).join(StockKeepingUnit.batches).where(StockKeepingUnit.product_id == id_)
        ).scalar_one()
        if count > 0:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail="Product has entries and cannot be deleted."
            )
        db.execute(delete(StockKeepingUnit).where(StockKeepingUnit.product_id == id_))
        db.execute(delete(Product).where(Product.id == id_))
        db.commit()
        return product_blank()


@router.get("", response_model=schemas.ProductBlank)
def show_blank(
    user: UserToken = Security(get_user, scopes=["products"]),
) -> schemas.ProductBlank:
    return product_blank()


@router.get("/list", response_model=List[schemas.Product])
def show_list(user: UserToken = Depends(get_user)) -> List[schemas.Product]:
    with SessionFuture() as db:
        return [
            product_info(item)
            for item in db.execute(
                select(Product)
                .order_by(desc(Product.is_active))
                .order_by(Product.product_group_id)
                .order_by(Product.name)
            )
            .scalars()
            .all()
        ]


@router.get("/query")
async def show_term(
    q: str = None,  # Query
    a: bool = None,  # Active
    p: bool = None,  # Is Purchased?
    e: bool = False,  # Extended
    s: bool = False,  # List separate SKUs
    v: Optional[uuid.UUID] = None,  # Vendor
    d: Optional[str] = None,  # Date
    current_user: UserToken = Depends(get_user),
):
    list_ = []
    with SessionFuture() as db:
        query_ = select(Product).join(Product.skus).options(contains_eager(Product.skus))
        if a is not None:
            query_ = query_.filter(Product.is_active == a)
        if p is not None:
            query_ = query_.filter(Product.is_purchased == p)
        if q is not None:
            for item in q.split():
                if item.strip() != "":
                    query_ = query_.filter(
                        or_(Product.name.ilike(f"%{item}%"), StockKeepingUnit.units.ilike(f"%{item}%"))
                    )
        query_ = query_.order_by(Product.name)

        for item in db.execute(query_).unique().scalars().all():
            skus = item.skus if s else item.skus[:1]
            for sku in skus:
                rc_price = get_rc_price(item.id, d, v, db)
                list_.append(
                    {
                        "id": sku.id if s else item.id,
                        "name": item.name,
                        "price": sku.price if rc_price is None else rc_price,
                        "units": sku.units if s else "",
                        "fraction": sku.fraction,
                        "fractionUnits": sku.fraction_units,
                        "productYield": sku.product_yield,
                        "isSold": item.is_sold,
                        "salePrice": sku.sale_price,
                        "isRateContracted": False if rc_price is None else True,
                    }
                    if e
                    else {
                        "id": sku.id if s else item.id,
                        "name": f"{item.name} ({sku.units})" if s else item.name,
                        "price": sku.price if rc_price is None else rc_price,
                        "isRateContracted": False if rc_price is None else True,
                    }
                )
    return list_


@router.get("/{id_}", response_model=schemas.Product)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["products"]),
) -> schemas.Product:
    with SessionFuture() as db:
        item: Product = db.execute(select(Product).where(Product.id == id_)).scalar_one()
        return product_info(item)


def product_info(product: Product) -> schemas.Product:
    product = schemas.Product(
        id=product.id,
        code=product.code,
        name=product.name,
        skus=[
            schemas.StockKeepingUnit(
                id=sku.id,
                isDefault=sku.is_default,
                units=sku.units,
                fraction=sku.fraction,
                fractionUnits=sku.fraction_units,
                productYield=sku.product_yield,
                price=sku.price,
                salePrice=sku.sale_price,
            )
            for sku in product.skus
        ],
        isActive=product.is_active,
        isFixture=product.is_fixture,
        isPurchased=product.is_purchased,
        isSold=product.is_sold,
        productGroup=schemas.ProductGroupLink(id=product.product_group.id, name=product.product_group.name),
    )
    return product


def product_blank() -> schemas.ProductBlank:
    return schemas.ProductBlank(
        name="",
        skus=[],
        isActive=True,
        isPurchased=True,
        isSold=False,
        isFixture=False,
    )


def get_rc_price(id_: uuid.UUID, d: Optional[str], v: Optional[uuid.UUID], db: Session) -> Optional[Decimal]:
    if d is None or v is None:
        return None
    date_ = datetime.strptime(d, "%d-%b-%Y")
    contracts = select(RateContract.id).where(
        RateContract.vendor_id == v, RateContract.valid_from <= date_, RateContract.valid_till >= date_
    )
    rc_price = db.execute(
        select(RateContractItem.price).where(
            RateContractItem.sku_id == id_, RateContractItem.rate_contract_id.in_(contracts)
        )
    ).scalar_one_or_none()
    return rc_price
