import uuid

from datetime import date, timedelta
from typing import List

import brewman.schemas.recipe as schemas
import brewman.schemas.recipe_item

from fastapi import APIRouter, Depends, HTTPException, Request, Security, status
from sqlalchemy import desc, func, or_, select
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_finish_date, get_start_date, set_period
from ..db.session import SessionFuture
from ..models.cost_centre import CostCentre
from ..models.inventory import Inventory
from ..models.journal import Journal
from ..models.product import Product
from ..models.recipe import Recipe
from ..models.recipe_item import RecipeItem
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..schemas.user import UserToken


router = APIRouter()


# @router.post("/new")  # "Recipes"
@router.post("", response_model=schemas.Recipe)
def save(
    data: schemas.RecipeIn,
    request: Request,
    user: UserToken = Security(get_user),
) -> schemas.Recipe:
    with SessionFuture() as db:
        recipe_product: Product = db.execute(select(Product).where(Product.id == data.product.id_)).scalar_one()

        if data.valid_till < data.valid_from:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Valid Till cannot be less than valid from",
            )
        sale_price = 0
        recipe_cost = 0
        if len(data.items) == 0:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Recipe has no ingredients",
            )

        recipe = Recipe(
            product_id=data.product.id_,
            quantity=round(data.quantity, 2),
            sale_price=round(data.sale_price, 2),
            valid_from=data.valid_from,
            valid_till=data.valid_till,
        )
        for item in data.items:
            product: Product = db.execute(select(Product).where(Product.id == item.product.id_)).scalar_one()
            quantity = round(data.quantity, 2)
            if product.is_purchased:
                ingredient_cost = get_purchased_product_cost(product.id, data.valid_from, data.valid_till, db)
            else:
                ingredient_cost = get_sub_product_cost(product.id, data.valid_from, data.valid_till, db)
            cost_per_unit = ingredient_cost / (product.fraction * product.product_yield)
            recipe_cost += cost_per_unit * quantity
            recipe.recipe_items.append(RecipeItem(None, product.id, quantity, ingredient_cost))

        recipe.cost_price = round(recipe_cost / recipe.quantity, 2)
        if recipe_product.is_sold:
            recipe_product.sale_price = sale_price

        save_recipe(recipe, db)
        db.commit()
        set_period(data.valid_from, data.valid_till, request.session)
        return recipe_info(recipe)


def save_recipe(recipe: Recipe, db: Session):
    product: Product = db.execute(select(Product).where(Product.id == recipe.product_id)).scalar_one()
    product.price = recipe.cost_price
    update_old_rows(
        recipe.product_id,
        recipe.valid_from,
        recipe.valid_till,
        recipe.effective_from,
        db,
    )
    db.add(recipe)
    for item in recipe.recipe_items:
        item.recipe_id = recipe.id
        db.add(item)


def get_purchased_product_cost(product_id: uuid.UUID, start_date: date, finish_date: date, db: Session):
    quantity_sum = func.sum(Journal.debit * Inventory.quantity).label("quantity")
    amount_sum = func.sum(
        Journal.debit * Inventory.quantity * Inventory.rate * (1 + Inventory.tax) * (1 - Inventory.discount)
    ).label("amount")
    costing = db.execute(
        select(quantity_sum, amount_sum)
        .join(Product.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(
            Inventory.product_id == product_id,
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
            Voucher.voucher_type == VoucherType.ISSUE,
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
        .group_by(Product)
    ).one_or_none()
    if costing is None:
        product = db.execute(select(Product).where(Product.id == product_id)).scalar_one()
        return product.price
    else:
        quantity, amount = costing
        return amount / quantity


def get_sub_product_cost(product_id: uuid.UUID, start_date: date, finish_date: date, db: Session):
    product: Product = db.execute(select(Product).where(Product.id == product_id)).scalar_one()
    old_recipe = (
        db.execute(
            select(Recipe).where(
                Recipe.product_id == product_id,
                Recipe.effective_to == None,
                Recipe.valid_from == start_date,
                Recipe.valid_till == finish_date,
            )
        )
        .scalars()
        .one_or_none()
    )

    if old_recipe is not None:
        return old_recipe.cost_price

    old_recipe = (
        db.execute(
            select(Recipe)
            .where(Recipe.product_id == product_id, Recipe.effective_to == None)
            .order_by(desc(Recipe.valid_till))
        )
        .scalars()
        .one_or_none()
    )

    if old_recipe is None:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Ingredient XXX has no recipe, cannot determine cost",
        )

    recipe_cost = 0
    recipe = Recipe(
        product_id=product_id,
        quantity=old_recipe.quantity,
        sale_price=product.sale_price,
        valid_from=start_date,
        valid_till=finish_date,
    )
    for item in old_recipe.recipe_items:
        if item.product.is_purchased:
            ingredient_cost = get_purchased_product_cost(item.product_id, start_date, finish_date, db)
        else:
            ingredient_cost = get_sub_product_cost(item.product_id, start_date, finish_date, db)
        cost_per_unit = ingredient_cost / (item.product.fraction * item.product.product_yield)
        recipe_cost += cost_per_unit * item.quantity
        recipe.recipe_items.append(RecipeItem(None, item.product_id, item.quantity, ingredient_cost))

    recipe.cost_price = round(recipe_cost / old_recipe.quantity, 2)

    save_recipe(recipe, db)
    return recipe_cost


def update_old_rows(product_id, valid_from: date, valid_till: date, effective_date: date, db: Session):
    old = (
        db.execute(
            select(Recipe)
            .where(
                Recipe.product_id == product_id,
                Recipe.valid_from < valid_till,
                Recipe.valid_till > valid_from,
                or_(Recipe.effective_to == None, Recipe.effective_to >= effective_date),
            )
            .order_by(Recipe.effective_from)
        )
        .scalars()
        .all()
    )

    new_recipes = []

    for item in old:
        if item.valid_from < valid_from:
            recipe = Recipe(
                product_id=item.product_id,
                quantity=item.quantity,
                cost_price=item.cost_price,
                sale_price=item.sale_price,
                valid_from=item.valid_from,
                valid_till=valid_from - timedelta(days=1),
                effective_from=effective_date,
            )
            for ri in item.recipe_items:
                recipe.recipe_items.append(RecipeItem(None, ri.product_id, ri.quantity, ri.price))
            new_recipes.append(recipe)
        if item.valid_till > valid_till:
            recipe = Recipe(
                product_id=item.product_id,
                quantity=item.quantity,
                cost_price=item.cost_price,
                sale_price=item.sale_price,
                valid_from=valid_till + timedelta(days=1),
                valid_till=item.valid_till,
                effective_from=effective_date,
            )
            for ri in item.recipe_items:
                recipe.recipe_items.append(RecipeItem(None, ri.product_id, ri.quantity, ri.price))
            new_recipes.append(recipe)

        if item.effective_from == effective_date and item.effective_to is None:
            for ri in item.recipe_items:
                db.delete(ri)
            db.delete(item)
        else:
            item.effective_to = effective_date

    for recipe in new_recipes:
        db.add(recipe)
        for item in recipe.recipe_items:
            item.recipe_id = recipe.id
            db.add(item)


@router.delete("/{id_}", response_model=schemas.RecipeBlank)
def delete_route(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.RecipeBlank:
    with SessionFuture() as db:
        recipe: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()
        if len(recipe.product.recipes) > 1:
            db.delete(recipe)
        else:
            db.delete(recipe)
            db.delete(recipe.product)
        db.commit()
        return recipe_blank()


@router.get("", response_model=schemas.RecipeBlank)
def show_blank(
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.RecipeBlank:
    return recipe_blank()


@router.get("/list", response_model=List[schemas.Recipe])
async def show_list(
    user: UserToken = Depends(get_user),
) -> List[schemas.Recipe]:
    with SessionFuture() as db:
        list_: List[Recipe] = (
            db.execute(
                select(Recipe).join(Recipe.product).order_by(Recipe.product_id).order_by(desc(Recipe.valid_till))
            )
            .scalars()
            .all()
        )
        return [
            schemas.Recipe(
                id=item.id,
                product=schemas.ProductLink(id=item.product.id, name=item.product.name),
                quantity=item.quantity,
                costPrice=item.cost_price,
                salePrice=item.sale_price,
                isLocked=item.is_locked,
                notes="",
                validFrom=item.valid_from,
                validTill=item.valid_till,
                effectiveFrom=item.valid_from,
                effectiveTill=item.valid_till,
                items=[],
            )
            for item in list_
        ]


@router.get("/{id_}", response_model=schemas.Recipe)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.Recipe:
    with SessionFuture() as db:
        item: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()
        return recipe_info(item)


def recipe_info(recipe: Recipe) -> schemas.Recipe:
    return schemas.Recipe(
        id=recipe.id,
        product=schemas.ProductLink(
            id=recipe.product_id,
            name=recipe.product.name,
        ),
        quantity=recipe.quantity,
        salePrice=recipe.sale_price,
        costPrice=recipe.cost_price,
        validFrom=recipe.valid_from,
        validTill=recipe.valid_till,
        isLocked=recipe.is_locked,
        notes="",
        items=[
            brewman.schemas.recipe_item.RecipeItem(
                id=item.id,
                product=schemas.ProductLink(
                    id=item.product.id,
                    name=item.product.name,
                ),
                quantity=item.quantity,
                price=item.price,
            )
            for item in recipe.items
        ],
    )


def recipe_blank() -> schemas.RecipeBlank:
    return schemas.RecipeBlank(
        quantity=0,
        costPrice=0,
        salePrice=0,
        isLocked=False,
        notes="",
        items=[],
    )
