from datetime import date, datetime

import brewman.schemas.cash_flow as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import desc, func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.account_type import AccountType
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.CashFlow)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["cash-flow"]),
) -> schemas.CashFlow:
    return schemas.CashFlow(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        body=schemas.CashFlowBody(operating=[], investing=[], financing=[], details=[]),
        footer=None,
    )


@router.get("/data")
def report_data(
    request: Request,
    s: str = None,
    f: str = None,
    user: UserToken = Security(get_user, scopes=["cash-flow"]),
):
    with SessionFuture() as db:
        start_date = datetime.strptime(s, "%d-%b-%Y")
        finish_date = datetime.strptime(f, "%d-%b-%Y")
        body, footer = build_report(start_date, finish_date, db)
        set_period(s, f, request.session)
        return {
            "startDate": s,
            "finishDate": f,
            "body": body,
            "footer": footer,
        }


@router.get("/{id_}")
def report_id(
    id_: int,
    request: Request,
    s: str = None,
    f: str = None,
    user: UserToken = Security(get_user, scopes=["cash-flow"]),
):
    with SessionFuture() as db:
        start_date = datetime.strptime(s, "%d-%b-%Y")
        finish_date = datetime.strptime(f, "%d-%b-%Y")
        details, footer = build_report_id(id_, start_date, finish_date, db)
        set_period(s, f, request.session)
        return {
            "startDate": s,
            "finishDate": f,
            "body": {"details": details},
            "footer": footer,
        }


def build_report(start_date: date, finish_date: date, db: Session):
    sub_query = (
        select(Voucher.id)
        .join(Voucher.journals)
        .join(Journal.account)
        .where(
            AccountBase.type_id == select(AccountType.id).where(AccountType.name == "Cash").scalar_subquery(),
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
        )
    )

    query = db.execute(
        select(AccountType, func.sum(Journal.signed_amount))
        .join(Journal.voucher)
        .join(Journal.account)
        .join(AccountBase.type_)
        .where(
            Voucher.id.in_(sub_query),
            AccountBase.type_id != select(AccountType.id).where(AccountType.name == "Cash").scalar_subquery(),
        )
        .group_by(AccountType)
        .order_by(func.sum(Journal.signed_amount))
    ).all()

    total_amount = 0
    cf = {"operating": [], "investing": [], "financing": []}
    for account_type, amount in query:
        total_amount += amount * -1
        cf[account_type.cash_flow_classification.lower()].append(
            {
                "name": account_type.name,
                "url": ["/", "cash-flow", str(account_type.id)],
                "amount": amount * -1,
            }
        )

    opening = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date < start_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            AccountBase.type_id == select(AccountType.id).where(AccountType.name == "Cash"),
        )
    ).scalar()

    closing = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .join(Journal.account)
        .where(
            Voucher.date <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            AccountBase.type_id == select(AccountType.id).where(AccountType.name == "Cash"),
        )
    ).scalar()

    return (
        cf,
        [
            {
                "name": "Net increase in cash and cash equivalents",
                "amount": total_amount,
            },
            {
                "name": "Cash and cash equivalents at beginning of period",
                "amount": opening,
            },
            {"name": "Cash and cash equivalents at end of period", "amount": closing},
        ],
    )


def build_report_id(account_type: int, start_date: date, finish_date: date, db: Session):
    details = []
    sub_query = (
        select(Voucher.id)
        .join(Voucher.journals)
        .join(Journal.account)
        .where(
            AccountBase.type_id == select(AccountType.id).where(AccountType.name == "Cash").scalar_subquery(),
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
        )
    )

    query = db.execute(
        select(AccountBase, func.sum(Journal.signed_amount))
        .join(Journal, Voucher.journals)
        .join(AccountBase, Journal.account)
        .where(Voucher.id.in_(sub_query), AccountBase.type_id == account_type)
        .group_by(AccountBase)
        .order_by(desc(func.sum(Journal.amount)))
    ).all()

    total_amount = 0
    for account, amount in query:
        total_amount += amount * -1
        details.append(
            {
                "name": account.name,
                "url": ["/", "ledger", str(account.id)],
                "amount": amount * -1,
            }
        )

    return details, [{"name": "total", "amount": total_amount}]
