import uuid

from datetime import datetime
from decimal import Decimal
from typing import Optional

import brewman.schemas.voucher as output

from fastapi import APIRouter, HTTPException, Security, status
from sqlalchemy import and_, distinct, func, or_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_first_day
from ..db.session import SessionFuture
from ..models.account import Account
from ..models.account_base import AccountBase
from ..models.attendance import Attendance
from ..models.attendance_type import AttendanceType
from ..models.cost_centre import CostCentre
from ..models.db_image import DbImage
from ..models.employee import Employee
from ..models.inventory import Inventory
from ..models.journal import Journal
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..routers import get_lock_info
from ..schemas.user import UserToken


router = APIRouter()


@router.post("/post-voucher/{id_}", response_model=output.Voucher)
def post_voucher(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["post-vouchers"]),
):
    try:
        with SessionFuture() as db:
            voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            account_types = (
                db.execute(
                    select(distinct(AccountBase.type_id)).where(
                        AccountBase.id.in_([vj.account_id for vj in voucher.journals])
                    )
                )
                .scalars()
                .all()
            )
            allowed, message = get_lock_info([voucher.date], voucher.voucher_type, account_types, db)
            if not allowed:
                raise HTTPException(
                    status_code=status.HTTP_423_LOCKED,
                    detail=message,
                )
            voucher.posted = True
            voucher.poster_id = user.id_
            db.commit()
            return voucher_info(voucher, db)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def check_delete_permissions(voucher: Voucher, user: UserToken):
    voucher_type = voucher.voucher_type.name.replace("_", "-").lower()
    if voucher_type in ["payment", "receipt"]:
        voucher_type = "journal"
    if voucher.posted and "edit-posted-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit posted vouchers",
        )
    elif voucher.user_id != user.id_ and "edit-other-user's-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit other user's vouchers",
        )
    elif voucher_type not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=f"You are not allowed ({voucher.voucher_type.name.replace('_', ' ').title()}) vouchers",
        )


@router.delete("/delete/{id_}")
def delete_voucher(
    id_: uuid.UUID,
    user: UserToken = Security(get_user),
):
    with SessionFuture() as db:
        voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
        images = db.execute(select(DbImage).where(DbImage.resource_id == voucher.id)).scalars().all()
        check_delete_permissions(voucher, user)
        account_types = (
            db.execute(
                select(distinct(AccountBase.type_id)).where(
                    AccountBase.id.in_([vj.account_id for vj in voucher.journals])
                )
            )
            .scalars()
            .all()
        )
        allowed, message = get_lock_info([voucher.date], voucher.voucher_type, account_types, db)
        if not allowed:
            raise HTTPException(
                status_code=status.HTTP_423_LOCKED,
                detail=message,
            )
        json_voucher = voucher_info(voucher, db)
        batches_to_delete = []
        if voucher.voucher_type == VoucherType.ISSUE:
            for item in voucher.journals:
                if item.debit == 1:
                    destination = item.cost_centre_id
                else:
                    source = item.cost_centre_id
            if source == CostCentre.cost_centre_purchase():
                batch_consumed = True
            elif destination == CostCentre.cost_centre_purchase():
                batch_consumed = False
            else:
                batch_consumed = None

            if batch_consumed is None:
                pass
            elif batch_consumed:
                for item in voucher.inventories:
                    item.batch.quantity_remaining += item.quantity
            else:
                for item in voucher.inventories:
                    if item.batch.quantity_remaining < item.quantity:
                        raise HTTPException(
                            status_code=status.HTTP_423_LOCKED,
                            detail=f"Only {item.batch.quantity_remaining} of {item.product.name} remaining.\n"
                            f"So it cannot be deleted",
                        )
                    item.batch.quantity_remaining -= item.quantity
        elif voucher.voucher_type == VoucherType.PURCHASE:
            for item in voucher.inventories:
                uses = db.execute(
                    select(func.count(Inventory.id)).where(Inventory.batch_id == item.batch.id, Inventory.id != item.id)
                ).scalar()
                if uses > 0:
                    raise HTTPException(
                        status_code=status.HTTP_423_LOCKED,
                        detail=f"{item.product.name} has been issued and cannot be deleted",
                    )
                batches_to_delete.append(item.batch)
        elif voucher.voucher_type == VoucherType.PURCHASE_RETURN:
            for item in voucher.inventories:
                item.batch.quantity_remaining += item.quantity
        for b in batches_to_delete:
            db.delete(b)
        db.delete(voucher)
        for image in images:
            db.delete(image)
        db.commit()
        json_voucher["type"] = VoucherType[json_voucher["type"].replace(" ", "_").upper()]
        return blank_voucher(info=json_voucher, db=db)


def voucher_info(voucher, db: Session):
    json_voucher = {
        "id": voucher.id,
        "date": voucher.date.strftime("%d-%b-%Y"),
        "isStarred": voucher.is_starred,
        "type": voucher.voucher_type.name,
        "posted": voucher.posted,
        "narration": voucher.narration,
        "journals": [],
        "inventories": [],
        "employeeBenefits": [],
        "incentives": [],
        "incentive": 0,
        "files": [],
        "creationDate": voucher.creation_date.strftime("%d-%b-%Y %H:%M"),
        "lastEditDate": voucher.last_edit_date.strftime("%d-%b-%Y %H:%M"),
        "user": {"id": voucher.user.id, "name": voucher.user.name},
        "poster": voucher.poster.name if voucher.posted else "",
    }
    if voucher.reconcile_date is not None:
        json_voucher["reconcileDate"] = voucher.reconcile_date.strftime("%d-%b-%Y")
    if voucher.voucher_type == VoucherType.PURCHASE:
        item = [j for j in voucher.journals if j.debit == -1][0]
        json_voucher["vendor"] = {"id": item.account.id, "name": item.account.name}
    elif voucher.voucher_type == VoucherType.ISSUE:
        item = [j for j in voucher.journals if j.debit == -1][0]
        json_voucher["source"] = {"id": item.cost_centre_id, "name": ""}
        item = [j for j in voucher.journals if j.debit == 1][0]
        json_voucher["destination"] = {"id": item.cost_centre_id, "name": ""}
    if voucher.voucher_type == VoucherType.PURCHASE_RETURN:
        item = [j for j in voucher.journals if j.debit == 1][0]
        json_voucher["vendor"] = {"id": item.account.id, "name": item.account.name}
    else:
        for item in voucher.journals:
            json_voucher["journals"].append(
                {
                    "id": item.id,
                    "debit": item.debit,
                    "amount": item.amount,
                    "account": {"id": item.account.id, "name": item.account.name},
                    "costCentre": {"id": item.cost_centre_id},
                }
            )
    for item in voucher.employee_benefits:
        json_voucher["employeeBenefits"].append(
            {
                "grossSalary": item.gross_salary,
                "daysWorked": item.days_worked,
                "esiEmployee": item.esi_ee,
                "pfEmployee": item.pf_ee,
                "esiEmployer": item.esi_er,
                "pfEmployer": item.pf_er,
                "employee": {
                    "id": item.journal.account.id,
                    "name": item.journal.account.name,
                    "designation": item.journal.account.designation,
                    "costCentre": {
                        "id": item.journal.account.cost_centre.id,
                        "name": item.journal.account.cost_centre.name,
                    },
                },
            }
        )
    for item in voucher.incentives:
        employee = db.execute(select(Employee).where(Employee.id == item.journal.account_id)).scalar_one()
        json_voucher["incentives"].append(
            {
                "employeeId": item.journal.account_id,
                "name": item.journal.account.name,
                "designation": employee.designation,
                "department": item.journal.account.cost_centre.name,
                "daysWorked": item.days_worked,
                "points": item.points,
            }
        )
    if len(json_voucher["incentives"]) > 0:
        json_voucher["incentive"] = next(x.amount for x in voucher.journals if x.account_id == Account.incentive_id())
    for item in voucher.inventories:
        text = (
            f"{item.batch.sku.product.name} ({item.batch.sku.units}) {item.batch.quantity_remaining:.2f}@"
            f"{item.batch.rate:.2f} from {item.batch.name.strftime('%d-%b-%Y')}"
        )
        json_voucher["inventories"].append(
            {
                "id": item.id,
                "quantity": item.quantity,
                "rate": item.rate,
                "tax": item.tax,
                "discount": item.discount,
                "amount": item.amount,
                "batch": {
                    "id": item.batch.id,
                    "name": text,
                    "quantityRemaining": item.batch.quantity_remaining,
                    "tax": item.batch.tax,
                    "discount": item.batch.discount,
                    "rate": item.batch.rate,
                    "sku": {
                        "id": item.batch.sku.id,
                        "name": f"{item.batch.sku.product.name} ({item.batch.sku.units})"
                        if item.batch.sku.units
                        else item.batch.sku.product.name,
                    },
                },
            }
        )
    images = db.execute(select(DbImage).where(DbImage.resource_id == voucher.id)).scalars().all()
    for image in images:
        json_voucher["files"].append(
            {
                "id": image.id,
                "resized": f"/db-image/{image.id}/resized",
                "thumbnail": f"/db-image/{image.id}/thumbnail",
            }
        )
    return json_voucher


def blank_voucher(info, db: Session):
    if "type" not in info:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Voucher Type is null",
        )
    type_ = info["type"]
    if "date" not in info:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Date cannot be null",
        )
    json_voucher = {
        "type": type_.name,
        "date": info["date"],
        "isStarred": False,
        "posted": False,
        "narration": "",
        "journals": [],
        "inventories": [],
        "employeeBenefits": [],
    }
    if type_ == VoucherType.JOURNAL:
        pass
    elif type_ == VoucherType.PAYMENT:
        account = None
        if info and "account" in info and info["account"]:
            account = (
                db.execute(select(AccountBase).where(AccountBase.id == uuid.UUID(info["account"])))
                .scalars()
                .one_or_none()
            )
        if account is not None:
            account = {"id": account.id, "name": account.name}
        else:
            account = AccountBase.cash_in_hand()
        json_voucher["journals"].append({"account": account, "amount": 0, "debit": -1})
    elif type_ == VoucherType.RECEIPT:
        account = None
        if info and "account" in info and info["account"]:
            account = (
                db.execute(select(AccountBase).where(AccountBase.id == uuid.UUID(info["account"])))
                .scalars()
                .one_or_none()
            )
        if account is not None:
            account = {"id": account.id, "name": account.name}
        else:
            account = AccountBase.cash_in_hand()
        json_voucher["journals"].append({"account": account, "amount": 0, "debit": 1})
    elif type_ == VoucherType.PURCHASE:
        json_voucher["vendor"] = AccountBase.local_purchase()

    elif type_ == VoucherType.PURCHASE_RETURN:
        json_voucher["vendor"] = AccountBase.local_purchase()
    elif type_ == VoucherType.ISSUE:
        if "source" in info:
            json_voucher["source"] = {"id": info["source"]}
        else:
            json_voucher["source"] = {"id": CostCentre.cost_centre_purchase()}
        if "destination" in info:
            json_voucher["destination"] = {"id": info["destination"]}
        else:
            json_voucher["destination"] = {"id": CostCentre.cost_centre_kitchen()}
    elif type_ == VoucherType.EMPLOYEE_BENEFIT:
        json_voucher["employeeBenefits"] = []
    elif type_ == VoucherType.INCENTIVE:
        json_voucher["incentives"], json_voucher["incentive"] = incentive_employees(info["date"], db)
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'Voucher of type "{type_}" does not exist.',
        )
    json_voucher["files"] = []
    return json_voucher


def incentive_employees(date_, db: Session):
    date_ = datetime.strptime(date_, "%d-%b-%Y")
    start_date = get_first_day(date_)
    finish_date = date_
    details = []
    employees = (
        db.execute(
            select(Employee)
            .where(Employee.joining_date <= finish_date, or_(Employee.is_active, Employee.leaving_date >= start_date))
            .order_by(Employee.cost_centre_id, Employee.designation, Employee.name)
        )
        .scalars()
        .all()
    )
    for employee in employees:
        att = (
            db.execute(
                select(Attendance).where(
                    Attendance.employee_id == employee.id,
                    Attendance.date >= start_date,
                    Attendance.date <= finish_date,
                    Attendance.is_valid == True,  # noqa: E712
                )
            )
            .scalars()
            .all()
        )
        att = sum(map(lambda x: AttendanceType.by_id(x.attendance_type).value, att))
        details.append(
            {
                "employeeId": employee.id,
                "name": employee.name,
                "designation": employee.designation,
                "department": employee.cost_centre.name,
                "daysWorked": att,
                "points": employee.points,
            }
        )

    amount = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .where(
            Journal.account_id == Account.incentive_id(),
            Voucher.voucher_type != VoucherType.ISSUE,
            or_(
                Voucher.date <= finish_date,
                and_(
                    Voucher.date == finish_date,
                    Voucher.voucher_type != VoucherType.INCENTIVE,
                ),
            ),
        )
    ).scalar()
    amount = 0 if amount is None else amount * Decimal(0.9) * -1
    return details, amount


def check_voucher_edit_allowed(voucher: Voucher, user: UserToken):
    if voucher.posted and "edit-posted-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit posted vouchers",
        )
    elif voucher.user_id != user.id_ and "edit-other-user's-vouchers" not in user.permissions:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="You are not allowed to edit other user's vouchers",
        )


def get_batch_quantity(id_: uuid.UUID, voucher_id: Optional[uuid.UUID], db: Session) -> Decimal:
    query = (
        select(func.sum(Inventory.quantity * Journal.debit))
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .where(Inventory.batch_id == id_, Journal.cost_centre_id == CostCentre.cost_centre_purchase())
    )
    if voucher_id is not None:
        query = query.where(Voucher.id != voucher_id)
    return db.execute(query).scalar_one() or Decimal(0)
